"""
This **module** is the basic setting for the force field format of soft bond
"""
from ...helper import Molecule, set_global_alternative_names
from . import bond_base

bond_base.BondType.Add_Property({"from_AorB": int})


@Molecule.Set_Save_SPONGE_Input("bond_soft")
def write_bond(self):
    """
    This **function** is used to write SPONGE input file

    :param self: the Molecule instance
    :return: the string to write
    """
    bonds = []
    for bond in self.bonded_forces.get("bond_soft", []):
        order = list(range(2))
        if bond.k != 0:
            if self.atom_index[bond.atoms[order[0]]] > self.atom_index[bond.atoms[order[-1]]]:
                temp_order = order[::-1]
            else:
                temp_order = order
            bonds.append("%d %d %f %f %d" % (self.atom_index[bond.atoms[temp_order[0]]]
                                             , self.atom_index[bond.atoms[temp_order[1]]], bond.k, bond.b,
                                             bond.from_AorB))

    if bonds:
        towrite = "%d\n" % len(bonds)
        bonds.sort(key=lambda x: list(map(int, x.split()[:2])))
        towrite += "\n".join(bonds)

        return towrite
    return None

set_global_alternative_names()



"""
This **module** is used to process topology and conformations
"""
import os.path
from abc import ABC, abstractmethod

import numpy as np
from .helper import get_rotate_matrix, ResidueType, Molecule, Residue, set_global_alternative_names, Xdict, \
    GlobalSetting, set_attribute_alternative_names, Xopen, Xprint
from .build import save_sponge_input
from .load import load_coordinate
from .forcefield.special.min import save_min_bonded_parameters, do_not_save_min_bonded_parameters
from .mdrun import run


def impose_bond(molecule, atom1, atom2, length):
    """
    This **function** is used to impose the distance in ``molecule`` between ``atom1`` and ``atom2`` to ``length``

    usage example::

        import Xponge
        import Xponge.forcefield.amber.ff14sb
        mol = ALA*10
        Impose_Bond(mol, mol.residues[0].CA, mol.residues[0].C, 1.2)

    .. ATTENTION::

        `atom1` and `atom2` should be bonded if they are in one residue

    :param molecule: a ``Molecule`` instance
    :param atom1: the base atom, which will not change its coordinate
    :param atom2: the atom to change its coordinate to fit the length
    :param length: distance in the unit of angstrom
    :return: None
    """
    crd = molecule.get_atom_coordinates()
    _, atom2_friends = molecule.divide_into_two_parts(atom1, atom2)
    r0 = crd[molecule.atom_index[atom2]] - crd[molecule.atom_index[atom1]]
    l0 = np.linalg.norm(r0)
    if l0 == 0:
        crd[molecule.atom_index[atom2]] += (1 / 3) ** (0.5)
        r0 = crd[molecule.atom_index[atom2]] - crd[molecule.atom_index[atom1]]
        l0 = np.linalg.norm(r0)
    dr = (length / l0 - 1) * r0
    crd[atom2_friends] += dr
    for atom in molecule.atoms:
        i = molecule.atom_index[atom]
        atom.x = crd[i][0]
        atom.y = crd[i][1]
        atom.z = crd[i][2]


def impose_angle(molecule, atom1, atom2, atom3, angle):
    """
    This **function** is used to impose the angle in ``molecule`` between ``atom1``, ``atom2`` and ``atom3`` \
 to ``angle``.

    .. ATTENTION::

        The pairs of ``atom1`` - ``atom2`` and ``atom2`` - ``atom3``  should be bonded.

    :param molecule: a ``Molecule`` instance
    :param atom1: the base atom, which will not change its coordinate
    :param atom2: the base atom, which will not change its coordinate
    :param atom3: the atom to change its coordinate to fit the angle
    :param angle: angle in the unit of rad
    :return: None
    """
    crd = molecule.get_atom_coordinates()
    _, atom3_friends = molecule.divide_into_two_parts(atom2, atom3)
    r12 = crd[molecule.atom_index[atom1]] - crd[molecule.atom_index[atom2]]
    r23 = crd[molecule.atom_index[atom3]] - crd[molecule.atom_index[atom2]]
    angle0 = np.arccos(np.dot(r12, r23) / np.linalg.norm(r23) / np.linalg.norm(r12))
    delta_angle = angle - angle0
    crd[atom3_friends] = np.dot(crd[atom3_friends] - crd[molecule.atom_index[atom2]],
                                get_rotate_matrix(np.cross(r12, r23), delta_angle)) + crd[molecule.atom_index[atom2]]
    for atom in molecule.atoms:
        i = molecule.atom_index[atom]
        atom.x = crd[i][0]
        atom.y = crd[i][1]
        atom.z = crd[i][2]


def impose_dihedral(molecule, atom1, atom2, atom3, atom4, dihedral):
    """
    This **function** is used to impose the dihedral in ``molecule`` between ``atom1``, ``atom2``, ``atom3`` \
 and ``atom4`` to ``dihedral``.

    .. ATTENTION::

        The pairs of ``atom1`` - ``atom2``,  ``atom2`` - ``atom3``  and ``atom3`` - ``atom4`` should be bonded.

    :param molecule: a ``Molecule`` instance
    :param atom1: the base atom, which will not change its coordinate
    :param atom2: the base atom, which will not change its coordinate
    :param atom3: the atom to change its coordinate to fit the angle
    :param atom4: the atom to change its coordinate to fit the angle
    :param dihedral: dihedral angle in the unit of rad
    :return: None
    """
    crd = molecule.get_atom_coordinates()
    _, atom4_friends = molecule.divide_into_two_parts(atom2, atom3)
    r12 = crd[molecule.atom_index[atom1]] - crd[molecule.atom_index[atom2]]
    r23 = crd[molecule.atom_index[atom3]] - crd[molecule.atom_index[atom2]]
    r34 = crd[molecule.atom_index[atom3]] - crd[molecule.atom_index[atom4]]
    r12xr23 = np.cross(r12, r23)
    r34xr23 = np.cross(r34, r23)
    cos = np.dot(r12xr23, r34xr23) / np.linalg.norm(r12xr23) / np.linalg.norm(r34xr23)
    cos = max(-0.999999, min(cos, 0.999999))
    dihedral0 = np.arccos(cos)
    dihedral0 = np.pi - np.copysign(dihedral0, np.cross(r34xr23, r12xr23).dot(r23))
    delta_angle = dihedral - dihedral0
    crd[atom4_friends] = np.dot(crd[atom4_friends] - crd[molecule.atom_index[atom3]],
                                get_rotate_matrix(r23, delta_angle)) + crd[molecule.atom_index[atom3]]
    for atom in molecule.atoms:
        i = molecule.atom_index[atom]
        atom.x = crd[i][0]
        atom.y = crd[i][1]
        atom.z = crd[i][2]


def _add_solvent_box(molecule, new_molecule, molmin, molmax, solshape, distance, tolerance, solcrd):
    """

    :param molecule:
    :param new_molecule:
    :param molmin:
    :param molmax:
    :param solshape:
    :param distance:
    :param tolerance:
    :param solcrd:
    :return:
    """
    x0 = molmin[0] - solshape[0] - distance[0]
    while x0 < molmax[0] + distance[3] + solshape[0]:
        y0 = molmin[1] - solshape[1] - distance[1]
        while y0 < molmax[1] + distance[4] + solshape[1]:
            z0 = molmin[2] - solshape[2] - distance[2]
            while z0 < molmax[2] + distance[5] + solshape[2]:
                if (molmax[0] + tolerance + solshape[0] > x0 > molmin[0] - tolerance - solshape[0] and
                        molmax[1] + tolerance + solshape[1] > y0 > molmin[1] - tolerance - solshape[1] and
                        molmax[2] + tolerance + solshape[2] > z0 > molmin[2] - tolerance - solshape[2]):
                    z0 += solshape[2]
                    continue
                for atom in new_molecule.atoms:
                    i = new_molecule.atom_index[atom]
                    atom.x = solcrd[i][0] + x0
                    atom.y = solcrd[i][1] + y0
                    atom.z = solcrd[i][2] + z0
                molecule |= new_molecule
                z0 += solshape[2]
            y0 += solshape[1]
        x0 += solshape[0]


def add_solvent_box(molecule, solvent, distance, tolerance=3):
    """
    This **function** adds a box full of solvents to a molecule.
    This will be performed in place for a Molecule and out of place for a ResidueType.

    :param molecule: the molecule to add the box, either a ``Molecule`` or a ``ResidueType``
    :param solvent: the solvent molecule, either a ``Molecule`` or a ``ResidueType``
    :param distance: the distance between the ``molecule`` and the box in the unit of Angstrom. \
 This can be an ``int`` or a ``float``, and it can be also a list of them with the length 3 or 6, \
 which represents the 3 or 6 directions respectively.
    :param tolerance: the distance between two molecules. 3 for default.
    :return: the Molecule instance
    """
    if isinstance(distance, (float, int)):
        distance = [distance] * 6
    elif not isinstance(distance, list):
        raise Exception("parameter distance should be a list, an int or a float")

    if len(distance) == 3:
        distance = distance + distance
    elif len(distance) != 6:
        raise Exception("the length of parameter distance should be 3 or 6")

    if isinstance(molecule, ResidueType):
        new_molecule = Molecule(molecule.name)
        res_a = Residue(molecule)
        for atom in molecule.atoms:
            res_a.Add_Atom(atom)
        new_molecule.Add_Residue(res_a)
        molecule = new_molecule

    molcrd = molecule.get_atom_coordinates()
    molmin = np.min(molcrd, axis=0)
    molmax = np.max(molcrd, axis=0)
    if isinstance(solvent, ResidueType):
        new_molecule = Molecule(solvent.name)
        res_a = Residue(solvent)
        for atom in solvent.atoms:
            res_a.Add_Atom(atom)
        new_molecule.Add_Residue(res_a)
    else:
        new_molecule = solvent.deepcopy()
    solcrd = new_molecule.get_atom_coordinates()
    solmin = np.min(solcrd, axis=0)
    solmax = np.max(solcrd, axis=0)
    solshape = solmax - solmin + tolerance

    _add_solvent_box(molecule, new_molecule, molmin, molmax, solshape, distance, tolerance, solcrd)
    return molecule

def h_mass_repartition(molecules, repartition_mass=1.1, repartition_rate=3, exclude_residue_name="WAT"):
    """
    This **function** repartition the mass of light atoms to the connected heavy atoms. \
 This can help the simulation run with a time step of 4 fs.

    .. ATTENTION::

        Many functions use mass to guess the element, the mass repartition may cause error. So call this function \
 at the final step please unless you know what you are doing.

    :param molecules: a ``Molecule``
    :param repartition_mass: if the mass of the atom is not greater than this value, it will be seen as a light atom. \
 1.1 for default and in the unit of Dalton.
    :param repartition_rate: The mass of the light atom will multiplied by this value.
    :param exclude_residue_name: the residue name which will not do the repartition. "WAT" for default.
    :return: None
    """
    for res in molecules.residues:
        if res.name == exclude_residue_name:
            continue
        for atom in res.atoms:
            if atom.mass <= repartition_mass:
                connect_atoms = res.type.connectivity[res.type.name2atom(atom.name)]
                assert len(connect_atoms) == 1
                origin_mass = atom.mass
                atom.mass *= repartition_rate
                delta_mass = atom.mass - origin_mass
                for heavy_atom in connect_atoms:
                    res.name2atom(heavy_atom.name).mass -= delta_mass


def solvent_replace(molecule, select, toreplace, sort=True):
    """
    This **function** replaces the solvent to some other molecules randomly.

    usage example::

        import Xponge
        import Xponge.forcefield.amber.ff14sb
        import Xponge.forcefield.amber.tip3p
        mol = ALA*10
        Add_Solvent_Box(mol, WAT, 10)
        Solvent_Replace(mol, WAT, {K:10, CL:10})
        #Solvent_Replace(mol, lambda res:res.name == "WAT", {K:10, CL:10})

    :param molecule: a ``Molecule`` instance
    :param select: a **function** to decide which residues should be replaced, \
or a Residue, a ResidueType or a Molecule with only one Residue, \
which the residues to be replaced have the same name
    :param toreplace: a dict, which stores the mapping of molecules to replace and the number of molecules. \
 Every molecule should be a ``ResidueType``, a ``Residue`` or a ``Molecule`` with only one ``Residue``.
    :param sort: whether to sort the residues after replacing
    :return: None
    """
    solvents = []
    for_sort = Xdict()
    if not callable(select):
        if isinstance(select, Molecule):
            select = select.residues[0]
        resname = select.name
        # pylint: disable=unnecessary-lambda-assignment
        select = lambda res: res.name == resname
    for i, resi in enumerate(molecule.residues):
        if select(resi):
            solvents.append(i)
            for_sort[resi] = float("inf")
        else:
            for_sort[resi] = float("-inf")

    np.random.shuffle(solvents)
    count = 0
    for key, value in toreplace.items():
        assert isinstance(key, ResidueType) or (isinstance(key, Molecule) and len(key.residues) == 1)
        if isinstance(key, Molecule):
            key = key.residues[0].type

        tempi = solvents[count:count + value]
        count += value
        for i in tempi:
            new_residue = Residue(key)
            crd_o = [molecule.residues[i].atoms[0].x, molecule.residues[i].atoms[0].y, molecule.residues[i].atoms[0].z]
            crd0 = [key.atoms[0].x, key.atoms[0].y, key.atoms[0].z]
            for atom in key.atoms:
                new_residue.Add_Atom(atom, x=atom.x + crd_o[0] - crd0[0],
                                     y=atom.y + crd_o[1] - crd0[1], z=atom.z + crd_o[2] - crd0[2])
            molecule.residues[i] = new_residue
            for_sort[new_residue] = count
    if sort:
        molecule.residues.sort(key=lambda res: for_sort[res])


def main_axis_rotate(molecule, direction_long=None, direction_middle=None, direction_short=None):
    """
    This **function** rotates the main axis of the molecule to the desired direction

    :param molecule: a ``Molecule`` instance
    :param direction_long: a list of three ``int`` or ``float`` to represent the direction vector. \
The long main axis will rotate to this direction.
    :param direction_middle: a list of three ``int`` or ``float`` to represent the direction vector. \
The middle main axis will rotate to this direction.
    :param direction_short: a list of three ``int`` or ``float`` to represent the direction vector. \
The short main axis will rotate to this direction.
    :return: None
    """
    if direction_long is None:
        direction_long = [0, 0, 1]
    if direction_middle is None:
        direction_middle = [0, 1, 0]
    if direction_short is None:
        direction_short = [1, 0, 0]
    molcrd = molecule.get_atom_coordinates()
    eye = np.zeros((3, 3))
    mass_of_center = np.zeros(3)
    total_mass = 0
    for i, atom in enumerate(molecule.atoms):
        xi, yi, zi = molcrd[i]
        total_mass += atom.mass
        mass_of_center += atom.mass * np.array([xi, yi, zi])
    mass_of_center /= total_mass

    for i, atom in enumerate(molecule.atoms):
        xi, yi, zi = molcrd[i] - mass_of_center
        eye += atom.mass * np.array([[yi * yi + zi * zi, -xi * yi, -xi * zi],
                                     [-xi * yi, xi * xi + zi * zi, -yi * zi],
                                     [-xi * zi, -yi * zi, xi * xi + yi * yi]])

    eigval, eigvec = np.linalg.eig(eye)
    t = np.argsort(eigval)
    matrix0 = np.vstack([direction_short, direction_middle, direction_long])
    rotation_matrix = np.dot(matrix0, np.linalg.inv(np.vstack((eigvec[:, t[2]], eigvec[:, t[1]], eigvec[:, t[0]]))))
    molcrd = np.dot(molcrd - mass_of_center, rotation_matrix) + mass_of_center
    for i, atom in enumerate(molecule.atoms):
        atom.x = molcrd[i][0]
        atom.y = molcrd[i][1]
        atom.z = molcrd[i][2]


def get_peptide_from_sequence(sequence, charged_terminal=True):
    """
    This **function** is used to get a peptide from the sequence

    :param sequence: a string, the serial
    :param charged_terminal: whether to change the terminal residues to the corresponding charged residue
    :return: a Molecule instance, the peptide
    """
    assert isinstance(sequence, str) and len(sequence) > 1
    temp_dict = Xdict({"A": "ALA", "G": "GLY", "V": "VAL", "L": "LEU", "I": "ILE", "P": "PRO",
                       "F": "PHE", "Y": "TYR", "W": "TRP", "S": "SER", "T": "THR", "C": "CYS",
                       "M": "MET", "N": "ASN", "Q": "GLN", "D": "ASP", "E": "GLU", "K": "LYS",
                       "R": "ARG", "H": "HIS"}, not_found_message="{} is not an abbreviation for an amino acid")
    temp_dict2 = Xdict({key: ResidueType.get_type(value) for key, value in temp_dict.items()},
                       not_found_message="{} is not an abbreviation for an amino acid")
    if charged_terminal:
        head = "N" + temp_dict[sequence[0]]
        tail = "C" + temp_dict[sequence[-1]]
    else:
        head = temp_dict[sequence[0]]
        tail = temp_dict[sequence[-1]]

    toret = ResidueType.get_type(head)

    for i in sequence[1:-1]:
        toret = toret + temp_dict2[i]
    toret += ResidueType.get_type(tail)
    return toret


def optimize(mol, step=2000, only_bad_coordinate=True, dt=1e-8, force_limit=50,
             epoch_limit=10, pbc=True, extra_commands=None):
    """
    This **function** is used to optimize the structure of the Molecule instance

    :param mol: the molecule to optimize
    :param step: the limited step for each epoch for minimization, 2000 for default
    :param only_bad_coordinate: whether to optimize all the atoms or the atoms whose coordinates are bad
    :param dt: the start dt for minimization
    :param force_limit: the minimization will stop if the largest difference in unit of `kcal/mol/A` \
between forces in two steps is not more than this value.
    :param epoch_limit: the minimization will stop if the epoch is not less than this value.
    :param pbc: whether to use the periodic box condition
    :param extra_commands: a dict, with the extra commands to pass to the MD engine
    :return: 0 for reaching the force limitation, 1 for reaching the epoch limitation
    """
    from tempfile import TemporaryDirectory
    Xprint("Optimizing", verbose=0)
    with TemporaryDirectory() as tempdir:
        temp_prefix = os.path.join(tempdir, "temp")
        temp_out = os.path.join(tempdir, "min")
        Xprint("    Parametering", verbose=0)
        save_min_bonded_parameters()
        if not pbc:
            box_length_backup = mol.box_length
            mol.box_length = [999, 999, 999]
        save_sponge_input(mol, temp_prefix)
        if not pbc:
            mol.box_length = box_length_backup
        do_not_save_min_bonded_parameters()
        temp_mdin_name = os.path.join(tempdir, "mdin.txt")
        mdin = Xopen(temp_mdin_name, "w")
        mdin.write(f"""temp
default_in_file_prefix = {temp_prefix}
rst = {temp_out}
crd = {temp_prefix}.dat
box = {temp_prefix}.box
mdout = {temp_out}.out
mdinfo = {temp_out}.info
mode = minimization
step_limit = {step}
write_information_interval = {step}
molecule_map_output  = 1
""")
        towrite = ""
        if extra_commands:
            for command, value in extra_commands.items():
                towrite += f"{command} = {value}\n"
        mdin.write(towrite)
        mdin.close()
        if pbc:
            all_to_use = f"SPONGE -mdin {temp_mdin_name} "
        else:
            all_to_use = f"SPONGE_NOPBC -mdin {temp_mdin_name} "
        if force_limit > 0:
            all_to_use += f"-frc {temp_out}.frc "
        if only_bad_coordinate:
            all_to_use += f"-mass_in_file {temp_prefix + '_fake_mass.txt'} "
        if GlobalSetting.verbose < 2:
            print_to = f" > {os.devnull}"
        else:
            print_to = ""
        Xprint("    Running", verbose=0)
        run(all_to_use + f"-dt {dt} {print_to}")
        all_to_use += f"-coordinate_in_file {temp_out+'_coordinate.txt'} "
        start_dt = dt
        if force_limit > 0:
            frc = np.fromfile(f"{temp_out}.frc", dtype=np.float32).reshape(-1, 3)
            frc = np.linalg.norm(frc, axis=1)
            last_frc = np.zeros_like(frc)
            epoch = 0
            Xprint("Epoch    Max Force Difference    Force Limit    Epoch Limit", verbose=0)
            while not np.all(np.abs(frc - last_frc) < force_limit) and epoch < epoch_limit:
                epoch += 1
                dt = np.max((np.min((1e-3, np.sqrt(1e-4 / np.max(frc)))), start_dt))
                run(all_to_use + f"-dt {dt} {print_to}")
                last_frc = frc
                frc = np.fromfile(f"{temp_out}.frc", dtype=np.float32).reshape(-1, 3)
                frc = np.linalg.norm(frc, axis=1)
                Xprint("{0:5d}{1:>24.4f}{2:15.4f}{3:15d}".format(epoch,
                                                                 np.max(np.abs(frc - last_frc)),
                                                                 force_limit,
                                                                 epoch_limit), verbose=0)
        load_coordinate(temp_out+'_coordinate.txt', mol)
        Xprint("Optimization Finished", verbose=0)
        if np.all(np.abs(frc - last_frc) < force_limit):
            return 0
        return 1


class Region(ABC):
    """
    This **abstract class** is used to define a region
    **New From 1.2.6.4**
    """
    def __init__(self):
        self.side = "in"
        self._side = True
        set_attribute_alternative_names(self)

    @abstractmethod
    def __contains__(self, item):
        pass

    def set_side(self, side):
        """
        This **function** is used to set the side of the region"

        :param side: either "in" or "out"
        :return: None
        """
        self.side = side
        if side == "in":
            self._side = True
        elif side == "out":
            self._side = False
        else:
            raise ValueError("side should be 'in' or 'out'")



class IntersectRegion:
    """
    This **class** is used to get the interset region of some regions
    **New From 1.2.6.4**

    :param *regions: the regions
    """
    def __init__(self, *regions):
        self.regions = regions

    def __contains__(self, item):
        for region in self.regions:
            if item not in region:
                return False
        return True


class UnionRegion:
    """
    This **class** is used to get the union region of some regions
    **New From 1.2.6.4**

    :param *regions: the regions
    """
    def __init__(self, *regions):
        self.regions = regions

    def __contains__(self, item):
        for region in self.regions:
            if item in region:
                return True
        return False


Region.register(UnionRegion)
Region.register(IntersectRegion)


class BlockRegion(Region):
    """
    This **class** is used to define a block region
    **New From 1.2.6.4**

    :param x_low: the lowest x coordinate of the block region
    :param y_low: the lowest y coordinate of the block region
    :param z_low: the lowest z coordinate of the block region
    :param x_high: the highest x coordinate of the block region
    :param y_high: the highest y coordinate of the block region
    :param z_high: the highest z coordinate of the block region
    :param side: either "in" or "out"
    """
    def __init__(self, x_low, y_low, z_low, x_high, y_high, z_high, side="in"):
        self.x_low = x_low
        self.y_low = y_low
        self.z_low = z_low
        self.x_high = x_high
        self.y_high = y_high
        self.z_high = z_high
        super().__init__()
        self.set_side(side)

    def __contains__(self, item):
        ans = self.x_low < item[0] < self.x_high and self.y_low < item[1] < self.y_high \
              and self.z_low < item[2] < self.z_high
        return ans if self._side else not ans


class SphereRegion(Region):
    """
    This **class** is used to define a sphere region
    **New From 1.2.6.4**

    :param x: the x coordinate of the sphere origin
    :param y: the y coordinate of the sphere origin
    :param z: the z coordinate of the sphere origin
    :param r: the radius of the sphere
    :param side: either "in" or "out"
    """
    def __init__(self, x, y, z, r, side="in"):
        self.x = x
        self.y = y
        self.z = z
        self._r2 = r * r
        super().__init__()
        self.set_side(side)

    def __contains__(self, item):
        ans = (item[0] - self.x) ** 2 + (item[1] - self.y) ** 2 + (item[2] - self.z) ** 2
        ans = ans < self._r2
        return ans if self._side else not ans


class FrustumRegion(Region):
    """
    This **class** is used to define a frustum region
    **New From 1.2.6.4**

    :param x1: the x coordinate of the first circle origin
    :param y1: the y coordinate of the first circle origin
    :param z1: the z coordinate of the first circle origin
    :param r1: the radius of the first circle origin
    :param x2: the x coordinate of the second circle origin
    :param y2: the y coordinate of the second circle origin
    :param z2: the z coordinate of the second circle origin
    :param r2: the radius of the second circle origin
    :param side: either "in" or "out"
    """
    def __init__(self, x1, y1, z1, r1, x2, y2, z2, r2, side="in"):
        self.r1 = r1
        self.r2 = r2
        self.o1 = np.array([x1, y1, z1], dtype=np.float32)
        self.o2 = np.array([x2, y2, z2], dtype=np.float32)
        self.axis = self.o2 - self.o1
        self.length = np.linalg.norm(self.axis)
        self.axis /= self.length
        self.k = (r2 - r1) / self.length
        super().__init__()
        self.set_side(side)

    def __contains__(self, item):
        crd = np.array(item) - self.o1
        length = np.linalg.norm(crd)
        projection = np.dot(crd, self.axis)
        distance = length * length - projection * projection
        r = self.r1 + self.k * projection
        ans = self.length > projection > 0 and distance < r * r
        return ans if self._side else not ans


class PrismRegion(Region):
    """
    This **class** is used to define a prism (parallelepiped) region
    **New From 1.2.6.4**

    :param x0: the x coordinate of the origin
    :param y0: the y coordinate of the origin
    :param z0: the z coordinate of the origin
    :param x1: the x coordinate of the first basis vector
    :param y1: the y coordinate of the first basis vector
    :param z1: the z coordinate of the first basis vector
    :param x2: the x coordinate of the second basis vector
    :param y2: the y coordinate of the second basis vector
    :param zz: the z coordinate of the second basis vector
    :param x3: the x coordinate of the third basis vector
    :param y3: the y coordinate of the third basis vector
    :param z3: the z coordinate of the third basis vector
    :param side: either "in" or "out"
    """
    def __init__(self, x0, y0, z0, x1, y1, z1, x2, y2, z2, x3, y3, z3, side="in"):
        self.l0 = np.array([x0, y0, z0], dtype=np.float32)
        self.l1 = np.array([x1, y1, z1], dtype=np.float32)
        self.l2 = np.array([x2, y2, z2], dtype=np.float32)
        self.l3 = np.array([x3, y3, z3], dtype=np.float32)
        self.n3 = np.cross(self.l1, self.l2)
        self.n3 /= np.linalg.norm(self.n3)
        self.n2 = np.cross(self.l3, self.l1)
        self.n2 /= np.linalg.norm(self.n2)
        self.n1 = np.cross(self.l2, self.l3)
        self.n1 /= np.linalg.norm(self.n1)
        self.length = np.array([np.dot(self.l1, self.n1), np.dot(self.l2, self.n2), np.dot(self.l3, self.n3)])
        assert np.all(self.length > 0), "The basis vectors should mmet the right-handed axis system requirements"
        super().__init__()
        self.set_side(side)

    def __contains__(self, item):
        crd = np.array(item) - self.l0
        ans = 0 <= np.dot(crd, self.n1) < self.length[0] and 0 <= np.dot(crd, self.n2) < self.length[1] and \
              0 <= np.dot(crd, self.n3) < self.length[2]

        return ans if self._side else not ans


class Lattice:
    """
    This **class** is used to help with the process of the lattice
    **New From 1.2.6.4**

    :param style: the style of the lattice. "custom", "template:NAME" or the key values in Lattice.styles
        "sc", "fcc", "bcc", "hcp", "diamond" are in the default template styles.
    :param basis_molecule: the unit to repeat, a Residue, a ResidueType or a Molecule
    :param scale: the scale of the lattice
    :param origin: a list with 3 numbers, the origin of the lattice. [0,0,0] for default.
    :param cell_length: the length of the unit cell. [1,1,1] for default.
    :param cell_angle: the angle of the unit cell. [90,90,90] for default.
    :param basis_position: a list of lists, every sublist has 3 numbers for coordinates.
    :param spacing: a list with 3 numbers, the spacing distance in x, y, z directions.
    """
    styles = Xdict(not_found_message="There is no lattice style named {}")

    def __init__(self, style="custom", basis_molecule=None, scale=None, origin=None, cell_length=None,
                 cell_angle=None, basis_position=None, spacing=None):
        self.basis_molecule = basis_molecule
        if style == "custom" or style.startswith("template:"):
            self.scale = 1
            self.origin = [0, 0, 0]
            self.cell_length = [1, 1, 1]
            self.cell_angle = np.array([90, 90, 90], dtype=float)
            self.spacing = [0, 0, 0]
            self.basis_position = []
            if self.scale is not None:
                self.scale = scale
            if origin is not None:
                self.origin = origin
            if cell_length is not None:
                self.cell_length = cell_length
            if cell_angle is not None:
                self.cell_angle = np.array(cell_angle)
            assert np.all(self.cell_angle <= 90), "the cell angle should be not greater than 90 degree"
            if spacing is not None:
                self.spacing = spacing
            if basis_position is not None:
                self.basis_position = basis_position
        else:
            old_style = Lattice.styles[style]
            self.scale = scale
            self.origin = old_style.origin
            self.cell_length = old_style.cell_length
            self.cell_angle = old_style.cell_angle
            self.basis_position = old_style.basis_position
            self.spacing = old_style.spacing
        if not style.startswith("template:") and self.basis_molecule is None:
            raise ValueError("basis molecule should not be None for a non-template lattice")
        if not style.startswith("template:") and self.scale is None:
            raise ValueError("scale should not be None for a non-template lattice")
        if style.startswith("template:"):
            style_name = style.split(":")[1].strip()
            Lattice.styles[style_name] = self

        set_attribute_alternative_names(self)

    @staticmethod
    def _judge_region(x1, y1, z1, x2, y2, z2, region, mol, basis_mol, res_len):
        """

        :param x1:
        :param y1:
        :param z1:
        :param x2:
        :param y2:
        :param z2:
        :param region:
        :param mol:
        :param basis_mol:
        :param res_len:
        :return:
        """
        if (x2, y2, z2) in region:
            mol |= basis_mol
            for res in mol.residues[res_len:]:
                for atom in res.atoms:
                    atom.x = atom.x - x1 + x2
                    atom.y = atom.y - y1 + y2
                    atom.z = atom.z - z1 + z2

    def create(self, box, region, mol=None):
        """
        This **function** is used to create basis molecules to the region in the box

        :param box: the box of the system
        :param region: the region to create the basis_molecule
        :param mol: if ``mol`` the Molecule instance is provided, basis molecules will be added to ``mol``
        :return: a new Molecule instance, or the Molecule instance ``mol``
        """
        if not isinstance(box, BlockRegion) or box.side == "out":
            raise ValueError("Box should only be a BlockRegion with side == 'in' !")
        if not mol:
            mol = Molecule("unnamed")
        mol.box_length = [box.x_high - box.x_low, box.y_high - box.y_low, box.z_high - box.z_low]
        cell_angle = self.cell_angle * np.pi / 180.0
        c0, c1, c2 = np.cos(cell_angle)
        _, _, s2 = np.sin(cell_angle)
        basis_mol = self.basis_molecule
        res_len = -1
        if isinstance(basis_mol, Molecule):
            basis_mol.atoms = [atom for residue in basis_mol.residues for atom in residue.atoms]
            res_len = -len(basis_mol.residues)
        dl1x = self.scale * (self.cell_length[0] + self.spacing[0])

        dl2x = self.scale * (self.cell_length[1] + self.spacing[1])
        dl2y = dl2x * s2
        dl2x *= c2

        dl3x = self.scale * (self.cell_length[2] + self.spacing[2])
        dl3y = dl3x * (c0 - c2 * c1) / s2
        dl3z = dl3x * np.sqrt(1 - c2 * c2 - c1 * c1 - c0 * c0 + 2 * c0 * c1 * c2) / s2
        dl3x *= c1
        bps = np.array([[dl1x * basis[0] + dl2x * basis[1] + dl3x * basis[2],
                         dl2y * basis[1] + dl3y * basis[2], dl3z * basis[2]] for basis in self.basis_position])
        x_init = 0
        y_init = 0
        z0 = box.z_low + self.origin[2]
        x1 = basis_mol.atoms[0].x
        y1 = basis_mol.atoms[0].y
        z1 = basis_mol.atoms[0].z
        while z0 < box.z_high:
            y0 = box.y_low + self.origin[1] + y_init
            while y0 < box.y_high:
                x0 = box.x_low + self.origin[0] + x_init
                while x0 < box.x_high:
                    for basis in bps:
                        x2 = basis[0] + x0
                        y2 = basis[1] + y0
                        z2 = basis[2] + z0
                        self._judge_region(x1, y1, z1, x2, y2, z2, region, mol, basis_mol, res_len)
                    x0 += dl1x
                x_init += dl2x
                x_init %= dl1x
                y0 += dl2y
            x_init += dl3x
            x_init %= dl1x
            y_init += dl3y
            y_init %= dl2y
            z0 += dl3z
        return mol


SIMPLE_CUBIC_LATTICE = Lattice("template:sc", basis_position=[[0, 0, 0]])

BODY_CENTERED_CUBIC_LATTICE = Lattice("template:bcc", basis_position=[[0, 0, 0], [0.5, 0.5, 0.5]])

FACE_CENTERED_CUBIC_LATTICE = Lattice("template:fcc", basis_position=[[0, 0, 0], [0.5, 0, 0.5],
                                                                      [0, 0.5, 0.5], [0.5, 0.5, 0]])

HEXAGONAL_CLOSE_PACKED_LATTICE = Lattice("template:hcp",
                                         basis_position=[[0, 0, 0], [1/3, 1/3, 0.5]],
                                         cell_angle=[90, 90, 60],
                                         cell_length=[1, 1, 2/3*np.sqrt(6)])

DIAMOND_LATTICE = Lattice("template:diamond", basis_position=[[0, 0, 0], [0, 0.5, 0.5],
                                                              [0.5, 0, 0.5], [0.5, 0.5, 0],
                                                              [0.25, 0.25, 0.25], [0.25, 0.75, 0.75],
                                                              [0.75, 0.25, 0.75], [0.75, 0.75, 0.25]])

set_global_alternative_names()



"""
This **module** is the basic setting for the force field property of charge
"""
import numpy as np
from ...helper import Molecule, AtomType, set_global_alternative_names, Xdict

AtomType.Add_Property({"charge": float})

AtomType.Set_Property_Unit("charge", "charge", "e")


@Molecule.Set_Save_SPONGE_Input("charge")
def write_charge(self):
    """
    This **function** is used to write SPONGE input file

    :param self: the Molecule instance
    :return: the string to write
    """
    towrite = "%d\n" % (len(self.atoms))
    towrite += "\n".join(["%.6f" % (atom.charge * 18.2223) for atom in self.atoms])
    return towrite


@Molecule.Set_MindSponge_Todo("charge")
def _do(self, sys_kwarg, ene_kwarg, use_pbc):
    """

    :return:
    """
    from mindsponge.potential import CoulombEnergy
    if "atom_charge" not in sys_kwarg:
        sys_kwarg["atom_charge"] = []
    sys_kwarg["atom_charge"].append([atom.charge for atom in self.atoms])
    if "charge" not in ene_kwarg:
        ene_kwarg["charge"] = Xdict()
        ene_kwarg["charge"]["function"] = lambda system, ene_kwarg: CoulombEnergy(
            atom_charge=system.atom_charge, length_unit='A', energy_unit='kcal/mol',
            use_pbc=use_pbc, use_pme=use_pbc,
            nfft=system.pbc_box.asnumpy().astype(int)//4*4,
            exclude_index=np.array(sys_kwarg["exclude"]))


set_global_alternative_names()



"""
This **package** sets the basic configuration of amber force field
"""
import os
from ...helper import set_global_alternative_names, Generate_New_Bonded_Force_Type
from ... import AtomType, load_parmdat, load_frcmod

from ..base import charge_base, mass_base, lj_base, bond_base, angle_base, dihedral_base, nb14_base,\
    virtual_atom_base, exclude_base

AMBER_DATA_DIR = os.path.dirname(__file__)

lj_base.LJType.combining_method_A = lj_base.Lorentz_Berthelot_For_A
lj_base.LJType.combining_method_B = lj_base.Lorentz_Berthelot_For_B

nb14_base.NB14Type.New_From_String(r"""
name    kLJ     kee
X-X     0.5     0.833333
""")

exclude_base.Exclude(4)

# pylint: disable=invalid-name
AmberCMapType = None

def load_parameters_from_parmdat(filename, prefix=True):
    """
    This **function** is used to get amber force field parameters from parmdat files

    :param filename: the name of the input file
    :param prefix: whether add the AMBER_DATA_DIR to the filename
    :return:
    """
    if prefix:
        filename = os.path.join(AMBER_DATA_DIR, filename)
    atoms, bonds, angles, propers, impropers, ljs, nb14s = load_parmdat(filename)
    AtomType.New_From_String(atoms)
    bond_base.BondType.New_From_String(bonds)
    angle_base.AngleType.New_From_String(angles)
    dihedral_base.ProperType.New_From_String(propers)
    dihedral_base.ImproperType.New_From_String(impropers)
    lj_base.LJType.New_From_String(ljs)
    nb14_base.NB14Type.New_From_String(nb14s)


def load_parameters_from_frcmod(filename, include_cmap=False, prefix=True):
    """
    This **function** is used to get amber force field parameters from frcmod files

    :param filename: the name of the input file
    :param include_cmap: whether include cmap
    :param prefix: whether add the AMBER_DATA_DIR to the filename
    :return: None
    """
    if prefix:
        filename = os.path.join(AMBER_DATA_DIR, filename)
    atoms, bonds, angles, propers, impropers, ljs, cmap = load_frcmod(filename)

    AtomType.New_From_String(atoms)
    bond_base.BondType.New_From_String(bonds)
    angle_base.AngleType.New_From_String(angles)
    dihedral_base.ProperType.New_From_String(propers)
    dihedral_base.ImproperType.New_From_String(impropers)
    lj_base.LJType.New_From_String(ljs)

    if include_cmap:
        global AmberCMapType
        if AmberCMapType is None:
            # pylint: disable=invalid-name
            AmberCMapType = Generate_New_Bonded_Force_Type("cmap", "1-2-3-4-5", {"resolution": int, "parameters": list},
                                                           False)

            @AmberCMapType.Type_Name_Getter
            def _(atoms):
                atom_names = [atom.type.name for atom in atoms]
                res_name = atoms[2].residue.name
                atom_names[2] = res_name + "@" + atom_names[2]
                return "-".join(atom_names)

            @AmberCMapType.Set_Same_Force_Function
            def cmap_same_force(_, atom_list):
                """
                This **function** is used to return the same force type for an atom list
                :param _:
                :param atom_list:
                :return:
                """
                return [atom_list]
        AmberCMapType.New_From_Dict(cmap)


set_global_alternative_names()



"""
This **module** set the generalized Born force field
"""
from ...helper import source, Xprint, Guess_Element_From_Mass, set_global_alternative_names

source("....")

AtomType.Add_Property({"GB_radii": float})
AtomType.Set_Property_Unit("GB_radii", "distance", "A")
AtomType.Add_Property({"GB_scaler": float})


def bondi_radii(atom):
    """
    This **function** receives an atom and sets the Bondi radii

    :param atom: the Atom instance
    :return: None
    """
    temp_dict = {"H": 1.2, "C": 1.7, "N": 1.55, "O": 1.52, "F": 1.47,
                 "P": 1.8, "S": 1.8, "Cl": 1.75, "Br": 1.85, "I": 1.98}
    element = Guess_Element_From_Mass(atom.mass)
    atom.GB_radii = temp_dict.get(element, 1.5)
    temp_dict2 = {"H": 0.85, "C": 0.72, "N": 0.79, "O": 0.85, "F": 0.88,
                  "P": 0.86, "S": 0.96}
    atom.GB_scaler = temp_dict2.get(element, 0.8)


def modified_bondi_radii(atom):
    """
    This **function** receives an atom and sets the modified Bondi radii

    :param atom: the Atom instance
    :return: None
    """
    temp_dict = {"H": 1.2, "C": 1.7, "N": 1.55, "O": 1.5, "F": 1.5,
                 "Si": 2.1, "P": 1.85, "S": 1.8, "Cl": 1.7, "Br": 1.85, "I": 1.98}
    element = Guess_Element_From_Mass(atom.mass)
    atom.GB_radii = temp_dict.get(element, 1.5)
    temp_dict2 = {"H": 0.85, "C": 0.72, "N": 0.79, "O": 0.85, "F": 0.88,
                  "P": 0.86, "S": 0.96}
    atom.GB_scaler = temp_dict2.get(element, 0.8)
    if element == "H":
        for atom_a in atom.residue.connectivity[atom]:
            element_a = Guess_Element_From_Mass(atom_a.mass)
            if element_a in ("C", "N"):
                atom.GB_radii = 1.3
            elif elementA in ('S', 'O', 'H'):
                atom.GB_radii = 0.8
            break


AVAILABLE_RADIUS_SET = {"bondi_radii": bondi_radii, "modified_bondi_radii": modified_bondi_radii}


def _show_reference(radius_set):
    """

    :param radius_set:
    :return:
    """
    # pylint: disable=comparison-with-callable
    if radius_set == bondi_radii:
        Xprint("""Reference for Bondi radii:
         A. Bondi
         van der Waals Volumes and Radii
         Journal of Physical Chemistry 1964 68 (3) 441-451
         DOI: 10.1021/j100785a001
     """)
    # pylint: disable=comparison-with-callable
    elif radius_set == modified_bondi_radii:
        Xprint("""Reference for modified Bondi radii:
    Vickie Tsui, David A. Case
    Theory and Applications of the Generalized Born Solvation Model in Macromolecular Simulations
    Biopolymers 2001 56 (4) 275-291
    DOI: 10.1002/1097-0282(2000)56:4<275::AID-BIP10024>3.0.CO;2-E
""")


def write_gb_radii_and_scaler(self):
    """
    This **function** is used to write gb radii and scaler when saving SPONGE inputs

    :param self: the Molecule instance
    :return: the string to write
    """
    towrite = "%d\n" % (len(self.atoms))
    towrite += "\n".join(["%.4f %.4f" % (atom.GB_radii, atom.GB_scaler) for atom in self.atoms])
    return towrite


def set_gb_radius(mol, radius_set=modified_bondi_radii):
    """
    This **function** is used to set GB radius for the molecule

    :param mol: the molecule, either a Molecule instance, a ResidueType instance or a Residue instance
    :param radius_set: a function, which receives an Atom instance and sets the radius
    :return: None
    """
    _show_reference(radius_set)
    if isinstance(mol, Molecule):
        for residue in mol.residues:
            for atom in residue.atoms:
                radius_set(atom)
    else:
        for atom in mol.atoms:
            radius_set(atom)
    mol.box_length = [999, 999, 999]
    Molecule.Set_Save_SPONGE_Input("gb")(write_gb_radii_and_scaler)


set_global_alternative_names()



"""
This **module** is the basic setting for the force field non bonded exclusion
"""
from functools import partial
from ... import Molecule
from ...helper import set_attribute_alternative_name, set_global_alternative_names, Xdict


class Exclude:
    """
    This **class** is used to set non bonded exclusion generally
    """

    #: The current effective Exclude class
    current = None


    def __init__(self, *args, **kwargs):
        n = 4
        if len(args) == 1:
            n = args[0]
        self.n = kwargs.get("n", n)
        Exclude.current = self
        set_attribute_alternative_name(self, self.get_excluded_atoms)

        def write_exclude(mol):
            exclude_numbers = 0
            excludes = []
            for atom in mol.atoms:
                temp = atom.extra_excluded_atoms.copy()
                atom_self_index = mol.atom_index[atom]
                filter_func = partial(lambda x, y: mol.atom_index[x] > y, y=atom_self_index)
                excludes.append(
                    list(map(lambda x: mol.atom_index[x], filter(filter_func, temp))))
                exclude_numbers += len(excludes[-1])
                for i in range(2, n + 1):
                    for aton in atom.linked_atoms.get(i, []):
                        if mol.atom_index[aton] > atom_self_index and aton not in temp:
                            temp.add(aton)
                            exclude_numbers += 1
                            excludes[-1].append(mol.atom_index[aton])
                for aton in atom.linked_atoms.get("v", []):
                    if mol.atom_index[aton] > atom_self_index and aton not in temp:
                        temp.add(aton)
                        exclude_numbers += 1
                        excludes[-1].append(mol.atom_index[aton])
                excludes[-1].sort()
            towrite = "%d %d\n" % (len(mol.atoms), exclude_numbers)
            for exclude in excludes:
                exclude.sort()
                towrite += "%d %s\n" % (len(exclude), " ".join([str(atom_index) for atom_index in exclude]))

            return towrite

        Molecule.Set_Save_SPONGE_Input("exclude")(write_exclude)

        # pylint: disable=unused-argument
        def todo(mol, sys_kwarg, ene_kwarg, use_pbc):
            if "exclude" not in ene_kwarg:
                ene_kwarg["exclude"] = []
            temp_dict = self.get_excluded_atoms(mol)
            temp_dict = {atom: list(set_) for atom, set_ in temp_dict.items()}
            exclude_max = max(len(atoms) for atoms in temp_dict.values())

            ene_kwarg["exclude"].append([[mol.atom_index[temp_dict[atom][j]]
                                          if j < len(temp_dict[atom]) else len(mol.atoms)
                                          for j in range(exclude_max)] for i, atom in enumerate(mol.atoms)])
            for li in ene_kwarg["exclude"][-1]:
                li.sort()


        Molecule.Set_MindSponge_Todo("exclude")(todo)

    def get_excluded_atoms(self, molecule):
        """
        This **function** gives the excluded atoms of a molecule

        :param molecule: a Molecule instance
        :return: a dict, which stores the atom - excluded atoms mapping
        """
        temp_dict = Xdict()
        for atom in molecule.atoms:
            temp_dict[atom] = atom.extra_excluded_atoms.copy()
            for i in range(2, self.n + 1):
                for aton in atom.linked_atoms.get(i, []):
                    temp_dict[atom].add(aton)
            for aton in atom.linked_atoms.get("v", []):
                temp_dict[atom].add(aton)
        return temp_dict

set_global_alternative_names()



"""
This **module** is the basic setting for the force field property of mass
"""
from ...helper import Molecule, AtomType, set_global_alternative_names

AtomType.Add_Property({"mass": float})


@Molecule.Set_Save_SPONGE_Input("mass")
def write_mass(self):
    """
    This **function** is used to write SPONGE input file

    :param self: the Molecule instance
    :return: the string to write
    """
    towrite = "%d\n" % (len(self.atoms))
    towrite += "\n".join(["%.3f" % (atom.mass) for atom in self.atoms])
    return towrite


#pylint: disable=unused-argument
@Molecule.Set_MindSponge_Todo("mass")
def _do_mass(self, sys_kwarg, ene_kwarg, use_pbc):
    """

    :return:
    """
    if "atom_mass" not in sys_kwarg:
        sys_kwarg["atom_mass"] = []
    sys_kwarg["atom_mass"].append([atom.mass for atom in self.atoms])

set_global_alternative_names()



"""
This **module** is the basic setting for the force field format of periodic proper and improper dihedral
"""
from itertools import permutations
from ... import Generate_New_Bonded_Force_Type
from ...helper import Molecule, set_global_alternative_names

# pylint: disable=invalid-name
ImproperType = Generate_New_Bonded_Force_Type("harmonic_improper", "1-3-2-3", {"k": float, "phi0": float}, False)

ImproperType.topology_matrix = [[1, 3, 2, 3],
                                [1, 1, 2, 3],
                                [1, 1, 1, 2],
                                [1, 1, 1, 1]]

ImproperType.Set_Property_Unit("k", "energy·angle^-2", "kcal/mol·rad^-2")
ImproperType.Set_Property_Unit("phi0", "angle", "rad")


@ImproperType.Set_Same_Force_Function
def improper_same_force(_, atom_list):
    """
    This **function** is used to return the same force type for an atom list
    :param _:
    :param atom_list:
    :return:
    """
    temp = []
    if isinstance(atom_list, str):
        atom_list_temp = [atom.strip() for atom in atom_list.split("-")]
        center_atom = atom_list_temp.pop(2)
        for atom_permutation in permutations(atom_list_temp):
            atom_permutation = list(atom_permutation)
            atom_permutation.insert(2, center_atom)
            temp.append("-".join(atom_permutation))
    else:
        atom_list_temp = list(atom_list)
        center_atom = atom_list_temp.pop(2)
        for atom_permutation in permutations(atom_list_temp):
            atom_permutation = list(atom_permutation)
            atom_permutation.insert(2, center_atom)
            temp.append(atom_permutation)
    return temp


@Molecule.Set_Save_SPONGE_Input("improper_dihedral")
def write_dihedral(self):
    """
    This **function** is used to write SPONGE input file

    :param self: the Molecule instance
    :return: the string to write
    """
    dihedrals = []

    for dihedral in self.bonded_forces.get("harmonic_improper", []):
        if dihedral.k != 0:
            temp_order = [2, 0, 1, 3]
            dihedrals.append("%d %d %d %d %f %f" % (self.atom_index[dihedral.atoms[temp_order[0]]],
                                                    self.atom_index[dihedral.atoms[temp_order[1]]],
                                                    self.atom_index[dihedral.atoms[temp_order[2]]],
                                                    self.atom_index[dihedral.atoms[temp_order[3]]], dihedral.k,
                                                    dihedral.phi0))

    if dihedrals:
        towrite = "%d\n" % len(dihedrals)
        dihedrals.sort(key=lambda x: list(map(float, x.split())))
        towrite += "\n".join(dihedrals)

        return towrite
    return None

set_global_alternative_names()



import setuptools
 
f = open("README.md", "r", encoding="utf-8")
long_description = f.read()
f.close()

#for formal
setuptools.setup(
    name="Xponge",
    version="1.2.6.12.0",
    author="Yijie Xia",  
    author_email="yijiexia@pku.edu.cn", 
    description="A Python package to perform pre- and post-processing of molecular simulations",
    long_description=long_description, 
    long_description_content_type="text/markdown",
    url="https://gitee.com/gao_hyp_xyj_admin/xponge",
    packages=setuptools.find_packages(),
    package_data = {"":['*.mol2', '*.frcmod', '*.dat', '*.itp']},
    install_requires = ["numpy"],
    entry_points = {
        "console_scripts": ["Xponge = Xponge.__main__:main", 
                            "Xponge.mdrun = Xponge.mdrun.__main__:main"] },
    classifiers=[
        "Programming Language :: Python :: 3 :: Only",
        'Development Status :: 5 - Production/Stable',
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6', 
)



"""
This **package** sets the protein configuration of charmm27 force field
"""
from ...helper import remove_real_global_variable, source, Xprint

source("....")
source("...charmm27")

load_mol2(os.path.join(CHARMM27_DATA_DIR, "protein.mol2"), as_template=True)

ResidueType.clear_type("HYP")
ResidueType.clear_type("CHYP")
ResidueType.clear_type("NHE")

remove_real_global_variable("HYP")
remove_real_global_variable("CHYP")
remove_real_global_variable("NHE")

residues = "ALA ARG ASN ASP CYS CYX GLN GLU GLY HID HIE HIP ILE LEU LYS MET PHE PRO SER THR TRP TYR VAL".split()

for resname in residues:
    res = ResidueType.get_type(resname)
    nres = ResidueType.get_type("N" + resname)
    cres = ResidueType.get_type("C" + resname)
    res.head, cres.head = "N", "N"
    res.head_length, cres.head_length = 1.3, 1.3
    res.head_next, cres.head_next = "CA", "CA"
    res.tail, nres.tail = "C", "C"
    res.tail_next, nres.tail_next = "CA", "CA"
    res.tail_length, nres.tail_length = 1.3, 1.3

    res.head_link_conditions.append({"atoms": ["CA", "N"], "parameter": 120 / 180 * np.pi})
    cres.head_link_conditions.append({"atoms": ["CA", "N"], "parameter": 120 / 180 * np.pi})

    if resname != "PRO":
        res.head_link_conditions.append({"atoms": ["H", "CA", "N"], "parameter": -np.pi})
        cres.head_link_conditions.append({"atoms": ["H", "CA", "N"], "parameter": -np.pi})
    else:
        res.head_link_conditions.append({"atoms": ["HA", "CA", "N"], "parameter": 0})
        cres.head_link_conditions.append({"atoms": ["HA", "CA", "N"], "parameter": 0})

    res.tail_link_conditions.append({"atoms": ["CA", "C"], "parameter": 120 / 180 * np.pi})
    res.tail_link_conditions.append({"atoms": ["O", "CA", "C"], "parameter": -np.pi})
    nres.tail_link_conditions.append({"atoms": ["CA", "C"], "parameter": 120 / 180 * np.pi})
    nres.tail_link_conditions.append({"atoms": ["O", "CA", "C"], "parameter": -np.pi})

    GlobalSetting.Add_PDB_Residue_Name_Mapping("head", resname, "N" + resname)
    GlobalSetting.Add_PDB_Residue_Name_Mapping("tail", resname, "C" + resname)

ResidueType.get_type("ACE").tail_next = "CH3"
ResidueType.get_type("ACE").tail_length = 1.3
ResidueType.get_type("ACE").tail_link_conditions.append({"atoms": ["CH3", "C"], "parameter": 120 / 180 * np.pi})
ResidueType.get_type("ACE").tail_link_conditions.append({"atoms": ["O", "CH3", "C"], "parameter": -np.pi})

ResidueType.get_type("NME").head_next = "CH3"
ResidueType.get_type("NME").head_length = 1.3
ResidueType.get_type("NME").head_link_conditions.append({"atoms": ["CH3", "N"], "parameter": 120 / 180 * np.pi})
ResidueType.get_type("NME").head_link_conditions.append({"atoms": ["H", "CH3", "N"], "parameter": -np.pi})

GlobalSetting.HISMap["DeltaH"] = "HD1"
GlobalSetting.HISMap["EpsilonH"] = "HE2"
GlobalSetting.HISMap["HIS"].update({"HIS": {"HID": "HID", "HIE": "HIE", "HIP": "HIP"},
                                    "CHIS": {"HID": "CHID", "HIE": "CHIE", "HIP": "CHIP"},
                                    "NHIS": {"HID": "NHID", "HIE": "NHIE", "HIP": "NHIP"}})

ResidueType.get_type("CYX").connect_atoms["ssbond"] = "SG"

Xprint("""Reference for protein of charmm27:
    MacKerell, Jr., A. D., Feig, M., Brooks, C.L., III, Extending the
    treatment of backbone energetics in protein force fields: limitations
    of gas-phase quantum mechanics in reproducing protein conformational
    distributions in molecular dynamics simulations, Journal of
    Computational Chemistry, 25: 1400-1415, 2004.

and 

    MacKerell, Jr., A. D.,  et al. All-atom
    empirical potential for molecular modeling and dynamics Studies of
    proteins.  Journal of Physical Chemistry B, 1998, 102, 3586-3616.

""")
# pylint:disable=undefined-variable



"""
This **module** is used to provide help functions and classes about namespace
"""
import sys
from inspect import currentframe
from importlib import import_module, reload
from types import MethodType, FunctionType
from functools import partial


def source(module, into_global=True, reload_module=False):
    """
    This **function** import the module and merge all the global variables into the caller module globals().
    In fact, this is similar to the python "import", but it is more convenient to relatively import and reload.

    usage example::

        #import numpy as np
        np = source("numpy", False)

        # from .. import forcefield import amber as ff
        # from ..forcefield.amber import *
        ff = source("..forcefield.amber")

    :param module: the module name to import
    :param into_global: whether to merge the global variables
    :param reload_module: whether to reload the module
    :return: the module to import
    """
    global_ = currentframe().f_back.f_globals
    module_ = import_module(module, package=global_["__name__"])
    if reload_module:
        reload(module_)
    if into_global:
        for key, value in module_.__dict__.items():
            if not key.startswith("_"):
                global_[key] = value
    return module_


# for the special alternative name
SPECIAL_STRINGS = {"Pdb": "PDB", "Sponge": "SPONGE", "Nb14": "NB14", "Lj": "LJ", "Gb": "GB", "Mindsponge": "MindSponge",
                   "Residuetype": "ResidueType", "Pubchem": "PubChem", "Resp": "RESP", "Name2atom": "Name2Atom"}


def set_real_global_variable(name, value):
    """
    This **function** is used to set the variable to real global variable

    usage::

        set_real_global_variable(hello, lambda: print(”hello“))
        hello()

    :param name: the name to the use as a global variable
    :param value: the value corresponding to the variable
    :return: None
    """
    if name[0].isdigit():
        name = "_" + name
    sys.modules.get("__main__").__dict__[name] = value


def remove_real_global_variable(name):
    """
    This **function** is used to remove the variable from real global variable

    usage::

        x = 1
        remove_real_global_variable("x")
        x
        # NameError: name 'x' is not defined

    :param name: the global variable to remove
    :return: None
    """
    sys.modules.get("__main__").__dict__.pop(name)


def set_alternative_name(obj, func, set_method):
    """
    This **function** is used to set the alternative names for a function to an object.
    This is the basic function, and some other partial functions with specific ``set_method`` is more useful.

    :param obj: the object
    :param func: the function
    :param set_method: the method to set
    :return: None
    """
    name = func.__name__
    set_method(obj, name, func)
    new_name = "_".join([i.capitalize() for i in name.split("_")])
    second_new_name = "".join([i.capitalize() for i in name.split("_")])
    third_new_name = second_new_name[0].lower() + second_new_name[1:]

    set_method(obj, new_name, func)
    set_method(obj, second_new_name, func)
    set_method(obj, third_new_name, func)
    new_new_name = new_name
    for t, newt in SPECIAL_STRINGS.items():
        new_new_name = new_new_name.replace(t, newt)
        second_new_name = second_new_name.replace(t, newt)
        third_new_name = third_new_name.replace(t, newt)
    if new_new_name != new_name:
        set_method(obj, new_new_name, func)
        set_method(obj, second_new_name, func)
        set_method(obj, third_new_name, func)


set_attribute_alternative_name = partial(set_alternative_name, set_method=setattr)


def set_classmethod_alternative_names(cls):
    """
    This **function** is used to set the attribute/method alternative names for a class
    usage example::

        class A:
            #@staticmethod
            @classmethod
            def hello_world(cls):
                print("hello")
        set_classmethod_alternative_names(A)
        A.hello_world()
        A.helloWorld()
        A.Hello_World()
        A.HelloWorld()

    :param cls: the class to set alternative names
    :return: None
    """
    types = list(cls.__bases__)
    types.append(cls)
    for sometype in types:
        old_dict = {}
        old_dict.update(sometype.__dict__)
        for i in old_dict:
            self_func = getattr(cls, i, None)
            if self_func and isinstance(self_func, (MethodType, FunctionType)) and not self_func.__name__.startswith(
                    "_"):
                set_attribute_alternative_name(cls, self_func)


def set_attribute_alternative_names(instance):
    """
    This **function** is used to set the attribute/method alternative names for an instance
    usage example::

        class A:
            def __init__(self):
                set_attribute_alternative_names(self)
            def hello_world():
                print("hello")
        a = A()
        a.hello_world()
        a.helloWorld()
        a.Hello_World()
        a.HelloWorld()

    :param instance: the instance to set names
    :return: None
    """
    types = list(type(instance).__bases__)
    types.append(type(instance))
    for sometype in types:
        for i in sometype.__dict__:
            self_func = getattr(instance, i, None)
            if self_func and isinstance(self_func, (MethodType, FunctionType)) and not self_func.__name__.startswith(
                    "_"):
                set_attribute_alternative_name(instance, self_func)


def _dict_set_method(obj, name, func):
    """

    :param obj:
    :param name:
    :param func:
    :return:
    """
    obj[name] = func


set_dict_value_alternative_name = partial(set_alternative_name, set_method=_dict_set_method)


def set_global_alternative_names(real_global=False):
    """
    This **function** is used to set the alternative names of the functions in the module to be global
    usage example::

        def hello_world():
            print("hello")
        set_global_alternative_names()
        hello_world()
        helloWorld()
        Hello_World()
        HelloWorld()

    :param real_global: make the variable to real global, which can be used without the module name
    :return: None
    """
    dic = currentframe().f_back.f_globals
    new_dict = {}
    for key, value in dic.items():
        if not isinstance(value, (FunctionType, type)) or value.__name__.startswith("_") or not key.islower():
            continue

        if real_global:
            if isinstance(value, FunctionType):
                def _global_set_method(obj, name, func):
                    obj[name] = func
                    set_real_global_variable(name, func)

                set_alternative_name(new_dict, value, _global_set_method)
            else:
                set_real_global_variable(value.__name__, value)
        else:
            set_dict_value_alternative_name(new_dict, value)
    dic.update(new_dict)

set_global_alternative_names()



"""
This **module** set the basic configuration for OL3 (RNA)
"""
from ...helper import source, Xprint

source("....")
amber = source("...amber")

amber.load_parameters_from_parmdat("parm10.dat")

load_mol2(os.path.join(AMBER_DATA_DIR, "RNA.mol2"), as_template=True)

for i in "AUCG":
    res = ResidueType.get_type(i)
    res5 = ResidueType.get_type(i + "5")
    res3 = ResidueType.get_type(i + "3")
    res.tail = "O3'"
    res.tail_next = "C3'"
    res5.tail = "O3'"
    res5.tail_next = "C3'"

    res.head = "P"
    res.head_next = "OP2"
    res3.head = "P"
    res3.head_next = "OP2"

    res.tail_link_conditions.append({"atoms": ["C3'", "O3'"], "parameter": 120 / 180 * np.pi})
    res.tail_link_conditions.append({"atoms": ["H3'", "C3'", "O3'"], "parameter": -54 / 180 * np.pi})
    res5.tail_link_conditions.append({"atoms": ["C3'", "O3'"], "parameter": 120 / 180 * np.pi})
    res5.tail_link_conditions.append({"atoms": ["H3'", "C3'", "O3'"], "parameter": -54 / 180 * np.pi})

    res.head_link_conditions.append({"atoms": ["OP2", "P"], "parameter": 108 / 180 * np.pi})
    res.head_link_conditions.append({"atoms": ["O5'", "OP2", "P"], "parameter": 113 / 180 * np.pi})
    res3.head_link_conditions.append({"atoms": ["OP2", "P"], "parameter": 108 / 180 * np.pi})
    res3.head_link_conditions.append({"atoms": ["O5'", "OP2", "P"], "parameter": 113 / 180 * np.pi})

    GlobalSetting.Add_PDB_Residue_Name_Mapping("head", res.name, res5.name)
    GlobalSetting.Add_PDB_Residue_Name_Mapping("tail", res.name, res3.name)

Xprint("""Reference for OL3:
  Marie Zgarbová, Michal Otyepka, Jiří Sponer, Arnošt Mládek, Pavel Banáš, Thomas E Cheatham 3rd, Petr Jurečka
    Refinement of the Cornell et al. Nucleic Acids Force Field Based on Reference Quantum Chemical Calculations of 
      Glycosidic Torsion Profiles.
    Journal of chemical theory and computation 2011 (7) 2886-2902
    DOI:10.1021/ct200162x
""")



"""
This **module** is the basic setting for the force field format of virtual atoms
"""
from ... import Generate_New_Bonded_Force_Type
from ...helper import Molecule, GlobalSetting, set_global_alternative_names

# pylint: disable=invalid-name
VirtualType2 = Generate_New_Bonded_Force_Type("vatom2", "1",
                                              {"atom0": int, "atom1": int, "atom2": int, "k1": float, "k2": float},
                                              False)
GlobalSetting.VirtualAtomTypes["vatom2"] = 3


@Molecule.Set_Save_SPONGE_Input("virtual_atom")
def write_virtual_atoms(self):
    """
    This **function** is used to write SPONGE input file

    :param self: the Molecule instance
    :return: the string to write
    """
    vatoms = []
    for vatom in self.bonded_forces.get("vatom2", []):
        vatoms.append("2 %d %d %d %d %f %f" % (self.atom_index[vatom.atoms[0]],
                                               self.atom_index[vatom.atoms[0]] + vatom.atom0,
                                               self.atom_index[vatom.atoms[0]] + vatom.atom1,
                                               self.atom_index[vatom.atoms[0]] + vatom.atom2,
                                               vatom.k1, vatom.k2))

    if vatoms:
        towrite = ""
        towrite += "\n".join(vatoms)

        return towrite
    return None

set_global_alternative_names()



"""
This **package** sets the basic configuration of charmm27 force field
"""
import os
from ... import GlobalSetting, load_ffitp, AtomType, ResidueType, set_global_alternative_names
from ..base import charge_base, mass_base, lj_base, bond_base, ub_angle_base, \
    dihedral_base, nb14_base, nb14_extra_base, improper_base, \
    virtual_atom_base, cmap_base, exclude_base

CHARMM27_DATA_DIR = os.path.dirname(__file__)

lj_base.LJType.combining_method_A = lj_base.Lorentz_Berthelot_For_A
lj_base.LJType.combining_method_B = lj_base.Lorentz_Berthelot_For_B

GlobalSetting.Set_Invisible_Bonded_Forces(["improper"])

dihedral_base.ProperType.New_From_String(r"""
name        k reset  phi0 periodicity
X-X-X-X     0 0      0    0
""")
exclude_base.Exclude(4)


def load_parameter_from_ffitp(filename, prefix=True):
    """
    This **function** is used to get amber force field parameters from GROMACS ffitp

    :param filename: the name of the input file
    :param prefix: whether add the CHARMM27_DATA_DIR to the filename
    :return: None
    """
    if prefix:
        filename = os.path.join(CHARMM27_DATA_DIR, filename)
    output = load_ffitp(filename)

    AtomType.New_From_String(output["atomtypes"])
    bond_base.BondType.New_From_String(output["bonds"])
    dihedral_base.ProperType.New_From_String(output["dihedrals"])
    lj_base.LJType.New_From_String(output["LJ"])
    ub_angle_base.UreyBradleyType.New_From_String(output["Urey-Bradley"])
    improper_base.ImproperType.New_From_String(output["impropers"])
    nb14_extra_base.NB14Type.New_From_String(output["nb14_extra"])
    nb14_base.NB14Type.New_From_String(output["nb14"])
    cmap_base.CMapType.New_From_Dict(output["cmaps"])


load_parameter_from_ffitp("forcefield.itp")

set_global_alternative_names()



"""
This **module** is used to calculate the RESP charge.
The **module** is not available on Windows.
"""
try:
    from pyscf import gto, scf
except ModuleNotFoundError as exc:
    raise ModuleNotFoundError(
        "To calculate RESP charge, 'pyscf' package needed. Maybe you need 'pip install pyscf'") from exc

import numpy as np
from ..helper import get_fibonacci_grid, Xprint, Xdict, set_global_alternative_names

# Pay Attention To !!!UNIT!!!
default_radius = {"H": 1.2, "C": 1.5, "N": 1.5,
                  "O": 1.4, "P": 1.8, "S": 1.75,
                  "F": 1.35, "Cl": 1.7, "Br": 2.3}


# Pay Attention To !!!UNIT!!!
def _get_mk_grid(assign, crd, area_density=1.0, layer=4, radius=None):
    """

    :param assign:
    :param crd:
    :param area_density:
    :param layer:
    :param radius:
    :return:
    """
    grids = []
    factor = area_density * 0.52918 * 0.52918 * 4 * np.pi
    real_radius = Xdict(not_found_message="Radius for element {} not found. Maybe you need to set a dict \
     ({element : radii}) to the argument variable radius when calling this function.")
    real_radius.update(default_radius)
    if radius:
        real_radius.update(radius)

    lists0 = np.array([1.4 + 0.2 * i for i in range(layer)])
    for i, atom in enumerate(assign.atoms):
        r0 = real_radius[atom] / 0.52918
        for r in r0 * lists0:
            grids.extend([*get_fibonacci_grid(int(factor * r * r), crd[i], r)])
    grids = np.array(grids).reshape(-1, 3)
    for i, atom in enumerate(assign.atoms):
        r0 = 1.39 * real_radius[atom] / 0.52918
        t = np.linalg.norm(grids - crd[i], axis=1)
        grids = grids[t >= r0, :]
    return grids


def _force_equivalence_q(q, extra_equivalence):
    """

    :param q:
    :param extra_equivalence:
    :return:
    """
    for eq_group in extra_equivalence:
        q_mean = np.mean(q[eq_group])
        q[eq_group] = q_mean
    return q


def _get_pyscf_mol(assign, basis, charge, spin, opt):
    """

    :param assign:
    :param basis:
    :param charge:
    :param spin:
    :param opt:
    :return:
    """
    mols = ""
    for i, atom in enumerate(assign.atoms):
        mols += "%s %f %f %f\n" % (atom, assign.coordinate[i][0], assign.coordinate[i][1], assign.coordinate[i][2])
    mol = gto.M(atom=mols, verbose=0, basis=basis, charge=charge, spin=spin)

    if spin == 0:
        fun = scf.RHF(mol)
    else:
        fun = scf.UHF(mol)

    if opt:
        from pyscf.geomopt.geometric_solver import optimize as geometric_opt
        mol = geometric_opt(fun)
        if spin == 0:
            fun = scf.RHF(mol)
        else:
            fun = scf.UHF(mol)
        assign.coordinate = mol.atom_coords() * 0.52918

    fun.run()
    return mol, fun


def _resp_scf_kernel(mol, assign, a, b, matrix_a, matrix_a0, matrix_b, q):
    """

    :param mol:
    :param assign:
    :param a:
    :param b:
    :param matrix_a:
    :param matrix_a0:
    :param matrix_b:
    :param q:
    :return:
    """
    step = 0
    q_last_step = q
    while step == 0 or np.max(np.abs(q - q_last_step)) > 1e-4:
        step += 1
        q_last_step = q
        for i in range(mol.natm):
            if assign.atoms[i] != "H":
                matrix_a[i][i] = matrix_a0[i][i] + a / np.sqrt(q_last_step[i] * q_last_step[i] + b * b)

        ainv = np.linalg.inv(matrix_a)
        q = np.dot(ainv, matrix_b)
        q = q[:-1]

    return q


def _find_tofit_second(mol, assign):
    """

    :param mol:
    :param assign:
    :return:
    """
    tofit_second = []
    fit_group = {i: -1 for i in range(mol.natm)}
    sublength = 0
    for i in range(mol.natm):
        if assign.Atom_Judge(i, "C4"):
            fit_group[i] = len(tofit_second)
            tofit_second.append([i])
            temp = []
            for j in assign.bonds[i].keys():
                if assign.atoms[j] == "H":
                    temp.append(j)
            if temp:
                for j in temp:
                    fit_group[j] = len(tofit_second)
                tofit_second.append(temp)
                sublength += len(temp) - 1

        if assign.Atom_Judge(i, "C3"):
            temp = []
            for j in assign.bonds[i].keys():
                if assign.atoms[j] == "H":
                    temp.append(j)
            if len(temp) == 2:
                fit_group[i] = len(tofit_second)
                tofit_second.append([i])
                for j in temp:
                    fit_group[j] = len(tofit_second)
                tofit_second.append(temp)
                sublength += 1
    return tofit_second, fit_group, sublength


def _correct_extra_equivalence(tofit_second, fit_group, sublength, extra_equivalence, atom_numbers):
    """

    :param tofit_second:
    :param fit_group:
    :param sublength:
    :param extra_equivalence:
    :param atom_numbers:
    :return:
    """
    if extra_equivalence:
        equi_group = [set() for i in extra_equivalence]
        for i, eq in enumerate(extra_equivalence):
            for eq_atom in eq:
                if fit_group[eq_atom] != -1:
                    equi_group[i].add(fit_group[eq_atom])
            equi_group[i] = list(equi_group[i])
            equi_group[i].sort()

        all_groups = set()
        for atom in range(atom_numbers):
            all_groups.add(fit_group[atom])
        all_groups_list = list(all_groups)
        all_groups_list.sort()

        group_map = {i: i for i in all_groups_list}
        for eq in equi_group:
            for group in eq:
                group_map[group] = eq[0]

        temp_max = 0
        for group in all_groups_list:
            if group == -1:
                continue
            if group_map[group] == group:
                group_map[group] = temp_max
                temp_max += 1
            else:
                group_map[group] = group_map[group_map[group]]

        temp = tofit_second
        tofit_second = [[] for i in range(temp_max)]
        for i, group in enumerate(temp):
            tofit_second[group_map[i]].extend(group)
            sublength -= len(group) - 1

        for group in tofit_second:
            sublength += len(group) - 1

        for atom in range(atom_numbers):
            fit_group[atom] = group_map[fit_group[atom]]

    return tofit_second, fit_group, sublength


def _get_a20_and_b20(total_length, tofit_second, fit_group, sublength, mol, matrix_a0, matrix_b, charge, q):
    """

    :param total_length:
    :param tofit_second:
    :param fit_group:
    :param sublength:
    :param mol:
    :param matrix_a0:
    :param matrix_b:
    :param charge:
    :param q:
    :return:
    """
    a20 = np.zeros((total_length, total_length))
    count = len(tofit_second)
    for i in range(mol.natm):
        if fit_group[i] == -1:
            fit_group[i] = count
            count += 1
        a20[mol.natm - sublength][fit_group[i]] += 1
        a20[fit_group[i]][mol.natm - sublength] += 1

    b20 = np.zeros(total_length)
    for i in range(mol.natm):
        b20[fit_group[i]] += matrix_b[i]
        for j in range(mol.natm):
            a20[fit_group[i]][fit_group[j]] += matrix_a0[i][j]

    b20[mol.natm - sublength] = charge
    count = 0
    for i in range(mol.natm):
        if fit_group[i] >= len(tofit_second):
            b20[mol.natm - sublength + count + 1] = q[i]
            a20[mol.natm - sublength + count + 1][len(tofit_second) + count] = 1
            a20[len(tofit_second) + count][mol.natm - sublength + count + 1] = 1
            count += 1
    return a20, b20


# Pay Attention To !!!UNIT!!!
def resp_fit(assign, basis="6-31g*", opt=False, charge=None, spin=0, extra_equivalence=None,
             grid_density=6, grid_cell_layer=4,
             radius=None, a1=0.0005, a2=0.001, two_stage=True, only_esp=False):
    """
    This **function** fits the RESP partial charge for an Assign instance
    The **function** is not available on Windows.

    :param assign: the Assign instance
    :param basis: the basis for Hartree-Fock calculation
    :param opt: whether do the geometry optimization
    :param charge: total charge of the molecule. If None, it will use the sum of the assign.charge
    :param spin: total spin of the molecule. ``S`` instead of ``2S+1``.
    :param extra_equivalence: the extra equivalence to constrain the charge
    :param grid_density: the density for grids to fit, in the unit of amgstrom^-3
    :param grid_cell_layer: the cell layer for grids to fit
    :param radius: the vdw radius for different elements. Default is ``default_radius`` in this module.
    :param a1: the restrain factor in the first step
    :param a2: the restrain factor in the second step
    :param two_stage: whether do the second stage fitting. If set to False, the second stage fitting will not be done
    :param only_esp: whether do the first stage fitting. If set to True, no restrained fitting will be done
    :return: a list of charges
    """
    if extra_equivalence is None:
        extra_equivalence = []
    if charge is None:
        charge = int(round(np.sum(assign.charge)))

    mol, fun = _get_pyscf_mol(assign, basis, charge, spin, opt)
    grids = _get_mk_grid(assign, mol.atom_coords(), grid_density, grid_cell_layer, radius)
    # step1
    # fit all atoms
    vnuc = 0
    matrix_a0 = np.zeros((mol.natm, mol.natm))
    for i in range(mol.natm):
        r = mol.atom_coord(i)
        z = mol.atom_charge(i)
        rp = r - grids
        for j in range(mol.natm):
            rpj = mol.atom_coord(j) - grids
            matrix_a0[i][j] = np.sum(1.0 / np.linalg.norm(rp, axis=1) / np.linalg.norm(rpj, axis=1))

        vnuc += z / np.einsum('xi,xi->x', rp, rp) ** .5

    matrix_a0 = np.hstack((matrix_a0, np.ones(mol.natm).reshape(-1, 1)))
    temp = np.ones(mol.natm + 1)
    temp[-1] = 0
    matrix_a0 = np.vstack((matrix_a0, temp.reshape(1, -1)))
    matrix_a = np.zeros_like(matrix_a0)
    matrix_a[:] = matrix_a0
    ainv = np.linalg.inv(matrix_a)

    try:
        from pyscf import df
        fakemol = gto.fakemol_for_charges(grids)
        vele = np.einsum('ijp,ij->p', df.incore.aux_e2(mol, fakemol), fun.make_rdm1())
    except MemoryError:
        dm = fun.make_rdm1()
        vele = []
        for p in grids:
            mol.set_rinv_orig_(p)
            vele.append(np.einsum('ij,ij', mol.intor('int1e_rinv'), dm))
        vele = np.array(vele)

    mep = vnuc - vele

    matrix_b = np.zeros((mol.natm + 1))
    for i in range(mol.natm):
        r = mol.atom_coord(i)
        rp = np.linalg.norm(r - grids, axis=1)
        matrix_b[i] = np.sum(mep / rp)
    matrix_b[-1] = charge
    matrix_b = matrix_b.reshape(-1, 1)

    q = np.dot(ainv, matrix_b.reshape(-1, 1))[:-1]

    if only_esp:
        return _force_equivalence_q(q, extra_equivalence)

    q = _resp_scf_kernel(mol, assign, a1, 0.1, matrix_a, matrix_a0, matrix_b, q)

    if not two_stage:
        return _force_equivalence_q(q, extra_equivalence)

    # step2
    # fit the sp3 C and the hydrogen connected to it (pay attention to the symmetry!)
    tofit_second, fit_group, sublength = _find_tofit_second(mol, assign)
    tofit_second, fit_group, sublength = _correct_extra_equivalence(tofit_second, fit_group, sublength,
                                                                    extra_equivalence, mol.natm)

    if tofit_second:
        total_length = mol.natm - sublength + 1 + mol.natm - sublength - len(tofit_second)

        a20, b20 = _get_a20_and_b20(total_length, tofit_second, fit_group, sublength, mol, matrix_a0, matrix_b, charge,
                                    q)

        matrix_a = np.zeros_like(a20)
        matrix_a[:] = a20[:]
        matrix_b = b20.reshape(-1, 1)
        ainv = np.linalg.inv(matrix_a)
        q_temp = np.dot(ainv, matrix_b)[:-1]

        a = a2
        b = 0.1
        step = 0
        q_last_step = q_temp
        while step == 0 or np.max(np.abs(q_temp - q_last_step)) > 1e-4:
            step += 1
            q_last_step = q_temp
            for i in range(mol.natm - sublength):
                if assign.atoms[i] != "H":
                    matrix_a[i][i] = a20[i][i] + a / np.sqrt(q_last_step[i] * q_last_step[i] + b * b)
            ainv = np.linalg.inv(matrix_a)
            q_temp = np.dot(ainv, matrix_b)[:-1]

        for i, group in enumerate(tofit_second):
            for j in group:
                q[j] = q_temp[i]

    return _force_equivalence_q(q, extra_equivalence)


set_global_alternative_names()

Xprint("""Reference for resp.py:
1. pyscf
  Q. Sun, T. C. Berkelbach, N. S. Blunt, G. H. Booth, S. Guo, Z. Li, J. Liu, J. McClain, S. Sharma, S. Wouters, and G. K.-L. Chan
    PySCF: the Python-based simulations of chemistry framework
    WIREs Computational Molecular Science 2018 8(e1340) 
    DOI: 10.1002/wcms.1340
    
2. ESP MK grid generation
  Brent H. Besler, Kenneth M. Merz Jr., Peter A. Kollman
    Atomic charges derived from semiempirical methods
    Journal of Computational Chemistry 1990 11 431-439
    DOI: 10.1002/jcc.540110404
    
3. RESP
  Christopher I. Bayly, Piotr Cieplak, Wendy Cornell, and Peter A. Kollman
    A well-behaved electrostatic potential-based method using charge restraints for deriving atomic char
    Journal of Physical Chemistry 1993 97(40) 10269-10280
    DOI: 10.1021/j100142a004
    
""")



"""
This **module** defines the terminal commands
"""
import argparse
from . import tools


def _mytest(subparsers):
    """

    :param subparsers:
    :return:
    """
    mytest = subparsers.add_parser("test", help="test the basic function of Xponge")
    mytest.add_argument("-o", metavar="test", default="test", help="the prefix for the output files")
    mytest.add_argument("-v", "--verbose", metavar="-1", default=-1, type=int,
                        help="the verbose level for output, 1 or -1")
    mytest.add_argument("-d", "--do", metavar="todo", nargs="*", action="append",
                        default=None, choices=["all", "base", "assign", "charmm27", "lattice", "fep"],
                        help="the unit tests need to do, should be 'all', \
or one or more of 'base', 'assign', 'charmm27', 'lattice', 'fep'")
    mytest.set_defaults(func=tools.test)


def _converter(subparsers):
    """

    :param subparsers:
    :return:
    """
    converter = subparsers.add_parser("converter", help="convert the format of coordinate file")
    converter.add_argument("-p", required=True, metavar="TOP", help="the name of the topology file")
    converter.add_argument("-c", metavar="CRD", help="the name of the coordinate file")
    converter.add_argument("-o", required=True, metavar="OUT", help="the name of the output file")
    converter.add_argument("-cf", metavar="GUESS", default="guess", choices=["guess", "sponge_crd", "sponge_traj"],
                           help='''the format of the topology file, can be "guess", "sponge_crd" or "sponge_traj"''')
    converter.add_argument("-of", metavar="GUESS", default="guess", choices=["guess", "sponge_crd", "sponge_traj"],
                           help='''the format of the output file, can be "guess", "sponge_crd" or "sponge_traj"''')
    converter.set_defaults(func=tools.converter)


def _maskgen(subparsers):
    """

    :param subparsers:
    :return:
    """
    maskgen = subparsers.add_parser("maskgen", help="""use VMD to generate a file to record the atom
 indexes of the corresponding mask""")
    maskgen.add_argument("-p", required=True, help="the topology file")
    maskgen.add_argument("-c", help="the coordinate file")
    maskgen.add_argument("-o", required=True, help="the output file")
    maskgen.add_argument("--vmd", metavar="vmd", default="vmd", help="the command to start vmd")
    maskgen.set_defaults(func=tools.maskgen)


def _exgen(subparsers):
    """

    :param subparsers:
    :return:
    """
    exgen = subparsers.add_parser("exgen",
                                  help='process bond-like, angle-like, dihedral-like files to get the atoms to exclude')
    exgen.add_argument('-n', type=int, required=True, help='the atom numbers')
    exgen.add_argument('-o', required=True, help='output exclude file name')
    exgen.add_argument('-b', '--bond', default=[], nargs='+', help='''bond-like input files: skip the first line,
 and there are 2 atoms in the head of following lines''')
    exgen.add_argument('-a', '--angle', default=[], nargs='+', help='''angle-like input files: skip the first line,
 and there are 3 atoms in the head of following lines''')
    exgen.add_argument('-d', '--dihedral', default=[], nargs='+', help='''dihedral-like input files:
 skip the first line, and there are 4 atoms in the head of following lines''')
    exgen.add_argument('-v', '--virtual', default=[], nargs='+', help='''virtual-atom-like input files:
 the first number indicates the virtual type''')
    exgen.add_argument('-e', '--exclude', default=[], nargs='+', help='''exclude-like input files:
 add the information of another exclude file''')
    exgen.set_defaults(func=tools.exgen)


def _name2name(subparsers):
    """

    :param subparsers:
    :return:
    """
    name2name = subparsers.add_parser("name2name",
                                      help="change the atom names of a residue from one file to another file")
    name2name.add_argument("-fformat", "-from_format", dest="from_format", choices=["mol2", "pdb", "gaff_mol2"],
                           required=True, help="the format of the file which is needed to change from")
    name2name.add_argument("-ffile", "-from_file", dest="from_file", required=True,
                           help="the name of the file which is needed to change from")
    name2name.add_argument("-fres", "-from_residue", dest="from_residue", default="",
                           help="the residue name in ffile if fformat == pdb")

    name2name.add_argument("-tformat", "-to_format", dest="to_format", choices=["mol2", "pdb", "gaff_mol2"],
                           required=True, help="the format of the file which is needed to change to")
    name2name.add_argument("-tfile", "-to_file", dest="to_file", required=True,
                           help="the name of the file which is needed to change to")
    name2name.add_argument("-tres", "-to_residue", dest="to_residue", default="",
                           help="the residue name in tfile if tformat == pdb")

    name2name.add_argument("-oformat", "-out_format", dest="out_format", choices=["mol2", "pdb", "mcs_pdb"],
                           required=True, help="the format of the output file")
    name2name.add_argument("-ofile", "-out_file", dest="out_file", required=True, help="the name of the output file")
    name2name.add_argument("-ores", "-out_residue", dest="out_residue", default="ASN",
                           help="the name of the output residue")
    name2name.add_argument("-tmcs", type=int, default=10, help="the time to find max common structure")
    name2name.set_defaults(func=tools.name2name)


def _mol2rfe(subparsers):
    """

    :param subparsers:
    :return:
    """
    mol2rfe = subparsers.add_parser("mol2rfe",
                                    help='calculate the relative free energy of a small molecule using SPONGE')
    mol2rfe.add_argument("-do", metavar="todo", nargs="*", action="append", help="""the things need to do,
 should be one or more of 'build', 'min', 'pre_equilibrium', 'equilibrium', 'analysis'""",
                         choices=["build", "min", "pre_equilibrium", "equilibrium", "analysis"])

    mol2rfe.add_argument("-pdb", required=True, help="the initial conformation given by the pdb file")
    mol2rfe.add_argument("-r2", "-residuetype2", required=True,
                         help="molecule mutated to by an Xponge ResidueType mol2 file")
    mol2rfe.add_argument("-r1", "-residuetype1", required=True,
                         help="molecule mutated from by an Xponge ResidueType mol2 file")
    mol2rfe.add_argument("-r0", "-residuetype0", nargs="*", default=[], help="small molecules that do not mutate")
    mol2rfe.add_argument("-ri", "-residue_index", type=int, metavar=0, default=0,
                         help="the residue index of the molecule to mutate")
    mol2rfe.add_argument("-nl", "-lambda_numbers", metavar=20, type=int, default=20,
                         help="the number of lambda groups - 1, default 20 for 0, 0.05, 0.10, 0.15..., 1.0")

    mol2rfe.add_argument("-dohmr", "-do_hydrogen_mass_repartition", action="store_true",
                         help="use the hydrogen mass repartition method")
    mol2rfe.add_argument("-ff", "-forcefield", help="Use this force field file instead of the default ff14SB and gaff")
    mol2rfe.add_argument("-mi", "-min_mdin", nargs="*", help="Use the minimization mdin file(s) here \
instead of the default ones")
    mol2rfe.add_argument("-pi", "-pre_equilibrium_mdin", help="Use this pre-equilibrium mdin file \
instead of the default one")
    mol2rfe.add_argument("-ei", "-equilibrium_mdin", help="Use this equilibrium mdin file instead of the default one")
    mol2rfe.add_argument("-ai", "-analysis_mdin", help="Use this analysis mdin file instead of the default one")
    mol2rfe.add_argument("-method", default="TI", choices=["TI"], help="the method to calculate the free energy")
    mol2rfe.add_argument("-temp", default="TMP", metavar="TMP", help="the temporary file name prefix")

    mol2rfe.add_argument("-tmcs", default=10, type=int, metavar="10",
                         help="the timeout parameter for max common structure in unit of second")
    mol2rfe.add_argument("-fmcs", default="mcs.png", type=str, metavar="mcs.png",
                         help="the file name for max common structure image")
    mol2rfe.add_argument("-lmcs", default=0.0, type=float, metavar="0.0",
                         help="minimum limitation of the Tanimoto coefficient of max common structure")
    mol2rfe.add_argument("-dt", default=2e-3, type=float, metavar="dt",
                         help="the dt used for simulation when mdin is not provided")
    mol2rfe.add_argument("-msteps", type=int, nargs=6,
                         help="""the minimization steps for all the lambda.
 Default 5000 for each minimization simulation. There are 6 minimization simulations.""",
                         default=[5000, 5000, 5000, 5000, 5000, 5000])
    mol2rfe.add_argument("-pstep", "-pre_equilibrium_step", dest="pre_equilibrium_step", default=50000, type=int,
                         metavar="pre_equilibrium_step",
                         help="the pre-equilibrium step used for simulation when mdin is not provided")
    mol2rfe.add_argument("-estep", "-equilibrium_step", dest="equilibrium_step", default=500000, type=int,
                         metavar="500000", help="the equilibrium step used for simulation when mdin is not provided")
    mol2rfe.add_argument("-thermostat", default="middle_langevin", metavar="middle_langevin",
                         help="the thermostat used for simulation when mdin is not provided")
    mol2rfe.add_argument("-barostat", default="andersen_barostat", metavar="andersen_barostat",
                         help="the barostat used for simulation when mdin is not provided")

    mol2rfe.set_defaults(func=tools.mol2rfe)


def main():
    """

    :return:
    """

    parser = argparse.ArgumentParser(prog="Xponge")
    subparsers = parser.add_subparsers(help="subcommands",
                                       description="Tools for SPONGE. Use Xponge XXX -h for the help of tool 'XXX'.")
    _mytest(subparsers)
    _maskgen(subparsers)
    _exgen(subparsers)
    _name2name(subparsers)
    _mol2rfe(subparsers)
    _converter(subparsers)

    args = parser.parse_args()

    if hasattr(args, "func"):
        args.func(args)
    else:
        parser.print_help()


if __name__ == "__main__":
    main()



"""
This **module** sets the residue types of D-furanose
"""
from itertools import product
from ....helper import source


def _init():
    """
    initialize the module
    """
    source("...glycam_06j")
    load_mol2(os.path.join(AMBER_DATA_DIR, "glycam_06j", "d_furanose.mol2"), as_template=True)
    all_types = ResidueType.get_all_types()
    for code, terminal, comformation in product("ABCDEFGHJKLMNOPQRSTUVWXYZ", "012346789PQRSTUVWXYZ", "UD"):
        resname = terminal + code + comformation
        if resname not in all_types:
            continue
        res = ResidueType.get_type(resname)
        tail_dihedral = 120 if comformation == "D" else -120
        if terminal == "1":
            res.head = "O1"
            res.head_next = "C1"
            res.head_length = 1.4
            res.head_link_conditions.append({"atoms": ["C1", "O1"], "parameter": 111 / 180 * np.pi})
            res.head_link_conditions.append({"atoms": ["O5", "C1", "O1"], "parameter": 60 / 180 * np.pi})
        elif code in "CPBJ":
            res.tail = "C2"
            res.tail_next = "C3"
            res.tail_length = 1.4
            res.tail_link_conditions.append({"atoms": ["O5", "C2"], "parameter": 110 / 180 * np.pi})
            res.tail_link_conditions.append({"atoms": ["C3", "O5", "C2"], "parameter": tail_dihedral / 180 * np.pi})
        else:
            res.tail = "C1"
            res.tail_next = "C2"
            res.tail_length = 1.4
            res.tail_link_conditions.append({"atoms": ["O4", "C1"], "parameter": 110 / 180 * np.pi})
            res.tail_link_conditions.append({"atoms": ["C2", "O4", "C1"], "parameter": tail_dihedral / 180 * np.pi})
        if terminal in "2ZYXTSP" or (terminal == "R" and code not in "VYW"):
            setHead(res, 2)
        elif terminal in "3WVQR":
            SetHead(res, 3)
        elif terminal in "4U":
            Set_Head(res, 4)
        elif terminal != "1":
            set_head(res, int(terminal))


_init()
